@@ -50,10 +50,9 @@ static PKCS11_CTX *ctx;
  * The memory for this PIN is always owned internally,
  * and may be freed as necessary. Before freeing, the PIN 
  * must be whitened, to prevent security holes.
- *
- * length is always MAX_PIN_LENGTH and possibly not 0 terminated?
  */
 static char *pin = NULL;
+static int pin_length = 0;
 
 static int verbose = 0;
 
@@ -91,6 +90,8 @@ int set_pin(const char *_pin)
 	/* Copy the PIN. If the string cannot be copied, NULL
 	   shall be returned and errno shall be set. */
 	pin = strdup(_pin);
+	if (pin != NULL)
+		pin_length = strlen(pin);
 
 	return (pin != NULL);
 }
@@ -113,11 +114,12 @@ static int get_pin(UI_METHOD * ui_method, void *callback_data)
 	} *mycb = callback_data;
 
 	/* pin in the call back data, copy and use */
-	if (mycb->password) {
+	if (mycb != NULL && mycb->password) {
 		pin = (char *)calloc(MAX_PIN_LENGTH, sizeof(char));
 		if (!pin)
 			return 0;
 		strncpy(pin,mycb->password,MAX_PIN_LENGTH);
+		pin_length = MAX_PIN_LENGTH;
 		return 1;
 	}
 
@@ -157,9 +159,10 @@ int pkcs11_finish(ENGINE * engine)
 		ctx = NULL;
 	}
 	if (pin != NULL) {
-		OPENSSL_cleanse(pin, MAX_PIN_LENGTH);
+		OPENSSL_cleanse(pin, pin_length);
 		free(pin);
 		pin = NULL;
+		pin_length = 0;
 	}
 	return 1;
 }
@@ -181,9 +184,10 @@ int pkcs11_init(ENGINE * engine)
 int pkcs11_rsa_finish(RSA * rsa)
 {
 	if (pin) {
-		OPENSSL_cleanse(pin, MAX_PIN_LENGTH);
+		OPENSSL_cleanse(pin, pin_length);
 		free(pin);
 		pin = NULL;
+		pin_length = 0;
 	}
 	if (module) {
 		free(module);
@@ -330,19 +334,10 @@ static int parse_slot_id_string(const char *slot_id, int *slot,
 
 	i = strspn(slot_id + 5, DIGITS);
 
-        /*
-         * changed 2008-01-28 9:31 GMT by SafeNet UK Ltd, akroehnert@safenet-inc.com
-         * changed comparision from == to >= as we need to check for any slot equal
-         * or greater zero
-         */
-	if (slot_id[i + 5] >= 0) {
+	if (slot_id[i + 5] == 0) {
 		*slot = n;
 		*id_len = 0;
-	/*
-	 * changed 2008-01-28 9:31 GMT by SafeNet UK Ltd, akroehnert@safenet-inc.com
-      * if we jump out here we cant get the label or id parsed
-	 */
-	/*	return 1; */
+		return 1; 
 	}
 
 	if (slot_id[i + 5] != '-') {
@@ -370,10 +365,7 @@ static int parse_slot_id_string(const char *slot_id, int *slot,
 
 	/* ... or "label_" */
 	if (strncmp(slot_id + i, "label_", 6) == 0)
-		/*
-	         * changed 2008-01-28 9:31 CET by SafeNet UK Ltd, akroehnert@safenet-inc.com
-        	 * we have to chop off + i + 6 characters instead of just 6
-		 */
+		*slot = n;
 		return (*label = strdup(slot_id + i + 6)) != NULL;
 
 	fprintf(stderr, "could not parse string!\n");
@@ -388,10 +380,11 @@ static int parse_slot_id_string(const char *slot_id, int *slot,
 static X509 *pkcs11_load_cert(ENGINE * e, const char *s_slot_cert_id)
 {
 	PKCS11_SLOT *slot_list, *slot;
+	PKCS11_SLOT *found_slot = NULL;
 	PKCS11_TOKEN *tok;
 	PKCS11_CERT *certs, *selected_cert = NULL;
 	X509 *x509;
-	unsigned int count, n, m;
+	unsigned int slot_count, cert_count, n, m;
 	unsigned char cert_id[MAX_VALUE_LEN / 2];
 	size_t cert_id_len = sizeof(cert_id);
 	char *cert_label = NULL;
@@ -425,14 +418,14 @@ static X509 *pkcs11_load_cert(ENGINE * e, const char *s_slot_cert_id)
 		}
 	}
 
-	if (PKCS11_enumerate_slots(ctx, &slot_list, &count) < 0)
+	if (PKCS11_enumerate_slots(ctx, &slot_list, &slot_count) < 0)
 		fail("failed to enumerate slots\n");
 
 	if (verbose) {
-		fprintf(stderr, "Found %u slot%s\n", count,
-			(count <= 1) ? "" : "s");
+		fprintf(stderr, "Found %u slot%s\n", slot_count,
+			(slot_count <= 1) ? "" : "s");
 	}
-	for (n = 0; n < count; n++) {
+	for (n = 0; n < slot_count; n++) {
 		slot = slot_list + n;
 		flags[0] = '\0';
 		if (slot->token) {
@@ -451,8 +444,14 @@ static X509 *pkcs11_load_cert(ENGINE * e, const char *s_slot_cert_id)
 			flags[m - 2] = '\0';
 		}
 
+		if (slot_nr != -1 &&
+			slot_nr == PKCS11_get_slotid_from_slot(slot)) {
+			found_slot = slot;
+		}
+
 		if (verbose) {
-			fprintf(stderr, "[%u] %-25.25s  %-16s", n,
+			fprintf(stderr, "[%lu] %-25.25s  %-16s",
+				PKCS11_get_slotid_from_slot(slot),
 				slot->description, flags);
 			if (slot->token) {
 				fprintf(stderr, "  (%s)",
@@ -464,20 +463,20 @@ static X509 *pkcs11_load_cert(ENGINE * e, const char *s_slot_cert_id)
 	}
 
 	if (slot_nr == -1) {
-		if (!(slot = PKCS11_find_token(ctx, slot_list, count)))
+		if (!(slot = PKCS11_find_token(ctx, slot_list, slot_count)))
 			fail("didn't find any tokens\n");
-	} else if (slot_nr >= 0 && slot_nr < count)
-		slot = slot_list + slot_nr;
-	else {
+	} else if (found_slot) {
+		slot = found_slot; 
+	} else {
 		fprintf(stderr, "Invalid slot number: %d\n", slot_nr);
-		PKCS11_release_all_slots(ctx, slot_list, count);
+		PKCS11_release_all_slots(ctx, slot_list, slot_count);
 		return NULL;
 	}
 	tok = slot->token;
 
 	if (tok == NULL) {
 		fprintf(stderr, "Found empty token; \n");
-		PKCS11_release_all_slots(ctx, slot_list, count);
+		PKCS11_release_all_slots(ctx, slot_list, slot_count);
 		return NULL;
 	}
 
@@ -486,18 +485,18 @@ static X509 *pkcs11_load_cert(ENGINE * e, const char *s_slot_cert_id)
 		fprintf(stderr, "Found token: %s\n", slot->token->label);
 	}
 
-	if (PKCS11_enumerate_certs(tok, &certs, &count)) {
+	if (PKCS11_enumerate_certs(tok, &certs, &cert_count)) {
 		fprintf(stderr, "unable to enumerate certificates\n");
-		PKCS11_release_all_slots(ctx, slot_list, count);
+		PKCS11_release_all_slots(ctx, slot_list, slot_count);
 		return NULL;
 	}
 
 	if (verbose) {
-		fprintf(stderr, "Found %u cert%s:\n", count,
-			(count <= 1) ? "" : "s");
+		fprintf(stderr, "Found %u cert%s:\n", cert_count,
+			(cert_count <= 1) ? "" : "s");
 	}
-	if ((s_slot_cert_id && *s_slot_cert_id) || (cert_id_len == 0)) {
-		for (n = 0; n < count; n++) {
+	if ((s_slot_cert_id && *s_slot_cert_id) && (cert_id_len != 0)) {
+		for (n = 0; n < cert_count; n++) {
 			PKCS11_CERT *k = certs + n;
 
 			if (cert_id_len != 0 && k->id_len == cert_id_len &&
@@ -511,7 +510,7 @@ static X509 *pkcs11_load_cert(ENGINE * e, const char *s_slot_cert_id)
 
 	if (selected_cert == NULL) {
 		fprintf(stderr, "certificate not found.\n");
-		PKCS11_release_all_slots(ctx, slot_list, count);
+		PKCS11_release_all_slots(ctx, slot_list, slot_count);
 		return NULL;
 	}
 
@@ -543,11 +542,12 @@ static EVP_PKEY *pkcs11_load_key(ENGINE * e, const char *s_slot_key_id,
 				 int isPrivate)
 {
 	PKCS11_SLOT *slot_list, *slot;
+	PKCS11_SLOT *found_slot = NULL;
 	PKCS11_TOKEN *tok;
 	PKCS11_KEY *keys, *selected_key = NULL;
 	PKCS11_CERT *certs;
 	EVP_PKEY *pk;
-	unsigned int count, n, m;
+	unsigned int slot_count, cert_count, key_count, n, m;
 	unsigned char key_id[MAX_VALUE_LEN / 2];
 	size_t key_id_len = sizeof(key_id);
 	char *key_label = NULL;
@@ -581,14 +581,14 @@ static EVP_PKEY *pkcs11_load_key(ENGINE * e, const char *s_slot_key_id,
 		}
 	}
 
-	if (PKCS11_enumerate_slots(ctx, &slot_list, &count) < 0)
+	if (PKCS11_enumerate_slots(ctx, &slot_list, &slot_count) < 0)
 		fail("failed to enumerate slots\n");
 
 	if (verbose) {
-		fprintf(stderr, "Found %u slot%s\n", count,
-			(count <= 1) ? "" : "s");
+		fprintf(stderr, "Found %u slot%s\n", slot_count,
+			(slot_count <= 1) ? "" : "s");
 	}
-	for (n = 0; n < count; n++) {
+	for (n = 0; n < slot_count; n++) {
 		slot = slot_list + n;
 		flags[0] = '\0';
 		if (slot->token) {
@@ -607,8 +607,14 @@ static EVP_PKEY *pkcs11_load_key(ENGINE * e, const char *s_slot_key_id,
 			flags[m - 2] = '\0';
 		}
 
+		if (slot_nr != -1 &&
+			slot_nr == PKCS11_get_slotid_from_slot(slot)) {
+			found_slot = slot;
+		}
+
 		if (verbose) {
-			fprintf(stderr, "[%u] %-25.25s  %-16s", n,
+			fprintf(stderr, "[%lu] %-25.25s  %-16s",
+				PKCS11_get_slotid_from_slot(slot),
 				slot->description, flags);
 			if (slot->token) {
 				fprintf(stderr, "  (%s)",
@@ -620,20 +626,20 @@ static EVP_PKEY *pkcs11_load_key(ENGINE * e, const char *s_slot_key_id,
 	}
 
 	if (slot_nr == -1) {
-		if (!(slot = PKCS11_find_token(ctx, slot_list, count)))
+		if (!(slot = PKCS11_find_token(ctx, slot_list, slot_count)))
 			fail("didn't find any tokens\n");
-	} else if (slot_nr >= 0 && slot_nr < count)
-		slot = slot_list + slot_nr;
-	else {
+	} else if (found_slot) {
+		slot = found_slot;
+	} else {
 		fprintf(stderr, "Invalid slot number: %d\n", slot_nr);
-		PKCS11_release_all_slots(ctx, slot_list, count);
+		PKCS11_release_all_slots(ctx, slot_list, slot_count);
 		return NULL;
 	}
 	tok = slot->token;
 
 	if (tok == NULL) {
 		fprintf(stderr, "Found empty token; \n");
-		PKCS11_release_all_slots(ctx, slot_list, count);
+		PKCS11_release_all_slots(ctx, slot_list, slot_count);
 		return NULL;
 	}
 /* Removed for interop with some other pkcs11 libs. */
@@ -645,7 +651,7 @@ static EVP_PKEY *pkcs11_load_key(ENGINE * e, const char *s_slot_key_id,
 #endif
 	if (isPrivate && !tok->userPinSet && !tok->readOnly) {
 		fprintf(stderr, "Found slot without user PIN\n");
-		PKCS11_release_all_slots(ctx, slot_list, count);
+		PKCS11_release_all_slots(ctx, slot_list, slot_count);
 		return NULL;
 	}
 
@@ -654,13 +660,13 @@ static EVP_PKEY *pkcs11_load_key(ENGINE * e, const char *s_slot_key_id,
 		fprintf(stderr, "Found token: %s\n", slot->token->label);
 	}
 
-	if (PKCS11_enumerate_certs(tok, &certs, &count))
+	if (PKCS11_enumerate_certs(tok, &certs, &cert_count))
 		fail("unable to enumerate certificates\n");
 
 	if (verbose) {
-		fprintf(stderr, "Found %u certificate%s:\n", count,
-			(count <= 1) ? "" : "s");
-		for (n = 0; n < count; n++) {
+		fprintf(stderr, "Found %u certificate%s:\n", cert_count,
+			(cert_count <= 1) ? "" : "s");
+		for (n = 0; n < cert_count; n++) {
 			PKCS11_CERT *c = certs + n;
 			char *dn = NULL;
 
@@ -685,19 +691,22 @@ static EVP_PKEY *pkcs11_load_key(ENGINE * e, const char *s_slot_key_id,
 			/* Free the PIN if it has already been 
 			   assigned (i.e, cached by get_pin) */
 			if (pin != NULL) {
-				OPENSSL_cleanse(pin, MAX_PIN_LENGTH);
+				OPENSSL_cleanse(pin, pin_length);
 				free(pin);
 				pin = NULL;
+				pin_length = 0;
 			}
 		} else if (pin == NULL) {
 			pin = (char *)calloc(MAX_PIN_LENGTH, sizeof(char));
+			pin_length = MAX_PIN_LENGTH;
 			if (pin == NULL) {
 				fail("Could not allocate memory for PIN");
 			}
 			if (!get_pin(ui_method, callback_data) ) {
-				OPENSSL_cleanse(pin, MAX_PIN_LENGTH);
+				OPENSSL_cleanse(pin, pin_length);
 				free(pin);
 				pin = NULL;
+				pin_length = 0;
 				fail("No pin code was entered");
 			}
 		}
@@ -706,9 +715,10 @@ static EVP_PKEY *pkcs11_load_key(ENGINE * e, const char *s_slot_key_id,
 		if (PKCS11_login(slot, 0, pin)) {
 			/* Login failed, so free the PIN if present */
 			if (pin != NULL) {
-				OPENSSL_cleanse(pin, MAX_PIN_LENGTH);
+				OPENSSL_cleanse(pin, pin_length);
 				free(pin);
 				pin = NULL;
+				pin_length = 0;
 			}
 			fail("Login failed\n");
 		}
@@ -730,19 +740,19 @@ static EVP_PKEY *pkcs11_load_key(ENGINE * e, const char *s_slot_key_id,
 	}
 
 	/* Make sure there is at least one private key on the token */
-	if (PKCS11_enumerate_keys(tok, &keys, &count)) {
+	if (PKCS11_enumerate_keys(tok, &keys, &key_count)) {
 		fail("unable to enumerate keys\n");
 	}
-	if (count == 0) {
+	if (key_count == 0) {
 		fail("No keys found.\n");
 	}
 
 	if (verbose) {
-		fprintf(stderr, "Found %u key%s:\n", count,
-			(count <= 1) ? "" : "s");
+		fprintf(stderr, "Found %u key%s:\n", key_count,
+			(key_count <= 1) ? "" : "s");
 	}
-	if (s_slot_key_id && *s_slot_key_id) {
-		for (n = 0; n < count; n++) {
+	if (s_slot_key_id && *s_slot_key_id && (key_id_len != 0 || key_label != NULL)) {
+		for (n = 0; n < key_count; n++) {
 			PKCS11_KEY *k = keys + n;
 
 			if (verbose) {
